package org.docx4j.samples;

import org.docx4j.openpackaging.io.LoadFromZipNG;
import org.docx4j.openpackaging.packages.OpcPackage;
import org.docx4j.openpackaging.parts.Part;
import org.docx4j.openpackaging.parts.relationships.RelationshipsPart;
import org.docx4j.relationships.Relationship;

import javax.imageio.ImageIO;
import javax.swing.*;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.image.BufferedImage;
import java.io.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import java.util.zip.ZipOutputStream;

/**
 * View a OpcPackage as a tree of it's zip entries, with a tree hierarchy
 * reflecting the OPC relationships.  Pretty print the XML, and allow
 * editing of the XML parts, which can then be saved back out.
 *
 * To-do:
 *  - save under different file name than the original
 *  - sort tree nodes alphabetically (hard to find stuff when there's a lot
 *    of nodes)
 *  - search the XML (simple string.indexOf("search term") will do)
 */
public class DocWin {

    // count # of open windows - when last window closes, exit
    private static int windowCount = 0;
    File file;
    JFrame frame;
    JSplitPane split;
    JTree tree;
    JScrollPane scroll;

    private Map<String,Object> data = new HashMap();
    private Map<String,JComponent> comps = new HashMap();

    DocWin(File f) throws Exception {
        file = f;
        OpcPackage pkg = new LoadFromZipNG().get(f);
        ZipFile zf = new ZipFile(file);
        frame = new JFrame(f.getName());

        DefaultMutableTreeNode root = new DefaultMutableTreeNode();
        root.add(populateTree(pkg.getRelationshipsPart(), zf, new HashMap()));
        // make sure to get [Content_Types].xml, as it has no part
        {
            root.add(new DefaultMutableTreeNode("[Content_Types].xml"));
            makeComponent(zf, "[Content_Types].xml");
        }
        tree = new JTree(root);
        tree.setRootVisible(false);
        tree.setShowsRootHandles(true);
        tree.addTreeSelectionListener(selListener);
        scroll = new JScrollPane();
        split = new JSplitPane();
        split.setLeftComponent(new JScrollPane(tree));
        split.setRightComponent(scroll);
        frame.getContentPane().add(split);
        //frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

        JMenuBar mb = new JMenuBar();
        JMenu menu = new JMenu("File");
        menu.add(makeBrowseItem());
        menu.add(makeSaveItem());
        mb.add(menu);
        frame.setJMenuBar(mb);

        // select root & expand it
        tree.setSelectionRow(0);
        tree.expandRow(0);
        treePathSelected("_rels/.rels");
        frame.setLocation(200,200);
        frame.setSize(500, 400);


        frame.addWindowListener(new WindowAdapter() {

            public void windowClosing(WindowEvent e) {
                if (--windowCount == 0) {
                    System.out.println("bye!");
                    System.out.flush();
                    System.exit(0);
                }
            }
        });

        windowCount++;
        frame.setVisible(true);
        split.setDividerLocation(.3d);
        zf.close();

    }

    private void treePathSelected(String sel) {
        JComponent comp = comps.get(sel);
        if (comp != null) {
            scroll.setViewportView(comp);
            scroll.repaint();
        } else {
            p("WARN: cannot find component for : " + sel);
        }
    }

    private TreeSelectionListener selListener = new TreeSelectionListener() {
        //@Override
        public void valueChanged(TreeSelectionEvent e) {
            if (e.getPath() != null) {
                treePathSelected(e.getPath().getLastPathComponent().toString());
            }
        }
    };
    private DefaultMutableTreeNode populateTree(RelationshipsPart relpart,
                                                ZipFile zf, Map handled)
            throws Exception {
        String pn = pn(relpart);
        DefaultMutableTreeNode ret = new DefaultMutableTreeNode(pn);
        makeComponent(zf, pn);
        // follow the UI convention of listing 'folder' nodes first,
        // and leaf nodes second
        java.util.List<Part> leafParts = new ArrayList();
        for (Relationship r : relpart.getRelationships().getRelationship()) {
            Part p = relpart.getPart(r);
            if (p == null) continue;
            pn = pn(p);
            if (handled.get(pn) != null) continue;
            handled.put(pn, true);
            if (p.getRelationshipsPart() == null) {
                leafParts.add(p);
                continue;
            }
            DefaultMutableTreeNode sub = new DefaultMutableTreeNode(pn);
            makeComponent(zf, pn);
            ret.add(sub);
            pn = pn(p.getRelationshipsPart());
            if (handled.get(pn) == null) {
                handled.put(pn, true);
                sub.add(populateTree(p.getRelationshipsPart(), zf, handled));
            }
        }
        // now add leaf nodes, those w/ no children
        for (Part p : leafParts) {
            pn = pn(p);
            DefaultMutableTreeNode sub = new DefaultMutableTreeNode(pn);
            makeComponent(zf, pn);
            ret.add(sub);
        }
        return ret;
    }
    private class SaveAction extends AbstractAction {
        public void actionPerformed(ActionEvent evt) {
            try {
                save();
            } catch (Exception e) {
                e.printStackTrace();
                System.exit(1);
            }
        }
    }
    private class OpenAction extends AbstractAction {
        public void actionPerformed(ActionEvent evt) {
            final JFileChooser chooser = new JFileChooser(file.getParentFile());
            chooser.setFileFilter(new javax.swing.filechooser.FileFilter() {
                public boolean accept(File f) {
                    if (f.isDirectory()) return true;
                    String n = f.getName().toLowerCase();
                    return n.endsWith(".docx") || n.endsWith(".pptx")
                            || n.endsWith(".xlsx");
                }
                public String getDescription() {
                    return "ooxml files";
                }
            });
            chooser.addActionListener(new AbstractAction() {
                public void actionPerformed(ActionEvent ev) {
                    chooser.setVisible(false);
                    File chosen = chooser.getSelectedFile();
                    if (chosen != null) {
                        try {
                            new DocWin(chosen);
                        } catch (Exception e) {
                            e.printStackTrace();
                        }
                    }
                };
            });
            chooser.showOpenDialog(DocWin.this.frame);
        }
    };
    JMenuItem makeBrowseItem() {
        JMenuItem ret = new JMenuItem("Open...");
        ret.addActionListener(new OpenAction());
        return ret;
    }
    JMenuItem makeSaveItem() {
        JMenuItem ret = new JMenuItem("Save");
        ret.addActionListener(new SaveAction());
        return ret;
    }
    JComponent makeComponent(ZipFile zf, String pn) throws Exception {
        JComponent ret;
        ZipEntry ze = zf.getEntry(pn);
        if (ze == null) throw new RuntimeException("cannot find '" + pn + "'");
        InputStream is = zf.getInputStream(ze);
        if (looksLikeXML(ze.getName()))
            ret = makeXMLComponent(ze, is);
        else if (looksLikeImage(ze.getName())) {
            data.put(ze.getName(), captureBytes(is));
            is.close();
            is = zf.getInputStream(ze);
            ret = makeImageComponent(ze, is);
        } else {
            data.put(ze.getName(), captureBytes(is));
            is.close();
            is = zf.getInputStream(ze);
            JTextArea ta = new JTextArea(loadText(is));
            ta.setEditable(false);
            ret = ta;
        }
        comps.put(pn, ret);
        return ret;
    }
    private void save() throws Exception {
        ZipOutputStream zos = new ZipOutputStream(new FileOutputStream(file));
        try {
            for (String fname : data.keySet()) {
                ZipEntry ze = new ZipEntry(fname);
                zos.putNextEntry(ze);
                Object o = data.get(fname);
                if (o instanceof byte[]) {
                    zos.write((byte[])o);
                } else {
                    String xml = ((JTextArea)o).getText();
                    xml = makeCompactXML(xml);
                    // fixme: i18n
                    zos.write(xml.getBytes("utf8"));
                }
                zos.closeEntry();
            }
            p("saved " + file.getAbsolutePath());
        } finally {
            zos.close();
        }
    }

    private byte[] captureBytes(InputStream is) throws IOException {
        byte[] buf = new byte[1024];
        int r;
        ByteArrayOutputStream bos = new ByteArrayOutputStream();
        while ((r = is.read(buf)) > 0)
            bos.write(buf, 0, r);
        is.close();
        return bos.toByteArray();
    }

    JComponent makeImageComponent(ZipEntry ze, InputStream is) throws Exception {
        return new ImagePanel(is);
    }
    JComponent makeXMLComponent(ZipEntry ze, InputStream is) throws Exception {
        String xml = loadPrettyXML(is);
        JTextArea ta = new JTextArea(xml);
        data.put(ze.getName(), ta);
        ta.setEditable(true);
        return ta;
    }

    public static void main(String[] args) throws Exception {
        if (args.length == 0) {
            System.err.println("error: no files specified");
            System.err.println("usage: DocWin [opc file 1] <opc-file-2> ...");
            System.exit(1);
        }
        for (String fileName : args) {
            File f = new File(fileName);
            new DocWin(f);
        }
    }

    private static boolean looksLikeImage(String name) {
        name = name.toLowerCase();
        return name.endsWith(".jpeg") || name.endsWith(".jpg")
                || name.endsWith(".gif") || name.endsWith(".png");
    }

    private String pn(Part p) {
        String ret = p.getPartName().toString();
        if (ret.charAt(0) == '/') ret = ret.substring(1);
        return ret;
    }

    /* XSLT to strip whitespace */
    private static final String XSLT = "<xsl:stylesheet version=\"1.0\"\n" +
            "   xmlns:xsl=\"http://www.w3.org/1999/XSL/Transform\">\n" +
            "\n" +
            "<xsl:output method=\"xml\" omit-xml-declaration=\"yes\"/>\n" +
            "\n" +
            "  <xsl:strip-space elements=\"*\"/>\n" +
            "\n" +
            "  <xsl:template match=\"@*|node()\">\n" +
            "   <xsl:copy>\n" +
            "    <xsl:apply-templates select=\"@*|node()\"/>\n" +
            "   </xsl:copy>\n" +
            "  </xsl:template>\n" +
            "\n" +
            "</xsl:stylesheet>";

    private String loadPrettyXML(InputStream is) throws Exception {
        // Instantiate transformer input

        Source xmlInput = new StreamSource(is);
        StreamResult xmlOutput = new StreamResult(new StringWriter());

        // Configure transformer
        Transformer transformer = TransformerFactory.newInstance()
                .newTransformer(); // An identity transformer
        transformer.setOutputProperty(OutputKeys.INDENT, "yes");
        transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "2");
        transformer.transform(xmlInput, xmlOutput);
        return xmlOutput.getWriter().toString();
    }

    // open office barfs on non-compact XML
    // one caveat: does not respect significant whitespace in run/text elements of document.xml
    private String makeCompactXML(String input) throws Exception {
        Source xmlInput = new StreamSource(new StringReader(input));

        // Configure transformer
        Transformer transformer = TransformerFactory.newInstance()
                .newTransformer(new StreamSource(new StringReader(XSLT)));
        transformer.setOutputProperty(OutputKeys.INDENT, "no");
        StreamResult xmlOutput = new StreamResult(new StringWriter());
        transformer.transform(xmlInput, xmlOutput);
        return xmlOutput.getWriter().toString();
    }

    private String loadText(InputStream is) throws Exception {
        Reader rdr = new InputStreamReader(is);
        StringBuffer sb = new StringBuffer();
        char[] buf = new char[1024];
        int r;
        while ((r = rdr.read(buf)) > 0) {
            sb.append(buf, 0, r);
        }
        return sb.toString();
    }
    static class ImagePanel extends JPanel{

        private BufferedImage image;

        public ImagePanel(InputStream is) throws Exception {
            image = ImageIO.read(is);
        }

        @Override
        public void paintComponent(Graphics g) {
            g.drawImage(image, 0, 0, null);

        }
    }
    static boolean looksLikeXML(String path) {
        path = path.toLowerCase();
        return path.endsWith(".xml") || path.endsWith(".rels");
    }
    private static void p(String s) {
        System.out.println("[DocWin] " + s);
    }

}

